package ca.nrc.naryrelations;

import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.apache.uima.UIMAException;
import org.apache.uima.analysis_engine.AnalysisEngine;
import org.apache.uima.analysis_engine.AnalysisEngineDescription;
import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.collection.CollectionReader;
import org.apache.uima.collection.CollectionReaderDescription;
import org.apache.uima.fit.component.JCasAnnotator_ImplBase;
import org.apache.uima.fit.descriptor.ConfigurationParameter;
import org.apache.uima.fit.factory.AnalysisEngineFactory;
import org.apache.uima.fit.factory.CollectionReaderFactory;
import org.apache.uima.fit.factory.JCasFactory;
import org.apache.uima.fit.pipeline.SimplePipeline;
import org.apache.uima.jcas.JCas;
import org.apache.uima.resource.ResourceInitializationException;
import org.apache.uima.util.CasCopier;
import org.dkpro.core.io.brat.BratReader;
import org.dkpro.core.io.brat.BratWriter;

public class BratUtils {
	
	public final static String BRAT_ANNOTATION_FILE_EXT = "ann";
	public final static String BRAT_TEXT_FILE_EXT = "txt";
	private final static String BRAT_UNKNOWN_LABEL = ".*";
	private final static String JAVA_UNKNOWN_TYPE = "de.tudarmstadt.ukp.dkpro.core.api.syntax.type.Tag";
	
	private static Map<String,JCas> capturedJCases = new HashMap<String,JCas>();
	
	public static void writeBratFile(JCas jcas, File bratFile) throws UIMAException, IOException, BratUtilsException {

		Logger tLogger = LogManager.getLogger("ca.nrc.nlp.brat.BratUtils.writeBratFile");
		tLogger.trace("Invoked with bratFile="+bratFile);
		traceBratFilesFor("At start of writeBratFile", bratFile, tLogger);

		File annFile = annFileFor(bratFile);
		File txtFile = txtFileFor(bratFile);
		
		// Write the .txt file
		{
			String text = jcas.getDocumentText();
			if (tLogger.isTraceEnabled()) tLogger.trace("Writing jcas text to to file="+txtFile+"\nWriting text=\n---------\n"+text+"\n---------\n\n");
			FileWriter writer = new FileWriter(txtFile);
			writer.write(text);
			writer.close();
			traceBratFilesFor("After writing .txt file", bratFile, tLogger);
		}
		
		// Write the .ann file
		{			
			// Make sure the .ann file exists
			if (!annFile.exists()) {
				new FileOutputStream(annFile).close();
			}
			
			AnalysisEngine engine = 
					AnalysisEngineFactory.createEngine(BratWriter.class,
							BratWriter.PARAM_TARGET_LOCATION, annFile.toString(),
						BratWriter.PARAM_SINGULAR_TARGET, true,
						BratWriter.PARAM_OVERWRITE, true
					);
			traceBratFilesFor("Just before running brat writing pipeline", bratFile, tLogger);
			
			SimplePipeline.runPipeline(jcas, new AnalysisEngine[] {engine});
			
			traceBratFilesFor("After running the Brat writing pipeline", bratFile, tLogger);
			
			removeDocContentFromAnnFile(jcas, bratFile);

			traceBratFilesFor("After removing document text fron .ann file", bratFile, tLogger);
		}
	}
	
	private static void removeDocContentFromAnnFile(JCas jcas, File bratFile) throws IOException {
		// For some unexplicable reason, BratWriter sometimes puts the content of the
		// document at the top of the .ann file and this causes an exception 
		// when you try to read the file back into a JCas.
		//
		
		String docText = jcas.getDocumentText();
		String annFileContent = annFileContent(bratFile);
		if (annFileContent.startsWith(docText)) {
			annFileContent = annFileContent.substring(docText.length());
			FileWriter writer = new FileWriter(annFileFor(bratFile));
			writer.write(annFileContent);
			writer.close();
		}
	}

	private static void traceBratFilesFor(String mess, File bratFile, Logger tLogger) throws IOException {
		if (tLogger.isTraceEnabled()) {
			if (mess == null) { mess = ""; }
			tLogger.trace(mess+": .txt file contains\n-----\n"+txtFileContent(bratFile)+"\n-----\n\n");
			tLogger.trace(mess+": .ann file contains\n-----\n"+annFileContent(bratFile)+"\n-----\n\n");
		}
	}

	private static File annFileFor(File bratFile) {
		File annFile = new File(bratFile.toString().replaceAll("\\.txt$", ".ann"));
		
		return annFile;
	}

	private static File txtFileFor(File bratFile) {
		File txtFile = new File(bratFile.toString().replaceAll("\\.ann$", ".txt"));
		
		return txtFile;
	}

	public static class JCasCapturer extends JCasAnnotator_ImplBase {
		
		public static final String PARAM_CAPTURE_ID = "id";
		@ConfigurationParameter(name = PARAM_CAPTURE_ID, mandatory = true)
		private String id;

		@Override
		public void process(JCas jcas) throws AnalysisEngineProcessException {
			
			JCas jcasCopy;
			try {
				jcasCopy = BratUtils.capturedJCasWithID(id);
			} catch (UIMAException e) {
				throw new AnalysisEngineProcessException(e);
			}
			
			jcasCopy.reset();
			CasCopier.copyCas(jcas.getCas(), jcasCopy.getCas(), true);

			return;
		}
	}
		
	public static JCas capturedJCasWithID(String id) throws UIMAException {
		if (!capturedJCases.containsKey(id)) {
			capturedJCases.put(id,  JCasFactory.createJCas());
		}
		return capturedJCases.get(id);
	}

	public static JCas readBratFile(File bratFile) throws UIMAException, IOException  {
		return readBratFile(bratFile, new String[][] {});
	}
	
	public static JCas readBratFile(File bratFile, String[][] brat2uimaMappings) throws UIMAException, IOException {
		Logger tLogger = LogManager.getLogger("ca.nrc.nlp.brat.BratUtils.readBratFile");

		if (tLogger.isTraceEnabled()) tLogger.trace("Invoked with bratFile="+bratFile);
		traceBratFilesFor("At start of readBratFile", bratFile, tLogger);
		
		String id = new Double(Math.random()).toString();
		
		CollectionReader reader = BratUtils.bratFileReader(bratFile, brat2uimaMappings);
		AnalysisEngine captureJCas = AnalysisEngineFactory.createEngine(
				BratUtils.JCasCapturer.class,
				BratUtils.JCasCapturer.PARAM_CAPTURE_ID, id);
		AnalysisEngine[] engines = new AnalysisEngine[] { captureJCas };
		
		SimplePipeline.runPipeline(reader, engines);
						
		JCas jcas = BratUtils.capturedJCasWithID(id);
		BratUtils.capturedJCases.remove(id);
		
		traceBratFilesFor("At END of readBratFile", bratFile, tLogger);

		return jcas;
	}
	
	private static String jcasContent(JCas jcas) {
		String content = "** JCas was null **";
		if (jcas != null) {
			content = jcas.getDocumentText();
		}
		return content;
	}
	
	private static String createBratJavaMapping(String leftType, String rightType) {
		return leftType + " -> " + rightType;
	}

	private static String createAnnotationType(String uimaType) {
		return uimaType + ":value";
	}
	
	public static CollectionReaderDescription bratDirReaderDescr(File dir, String lang, String[][] typeMappings) throws ResourceInitializationException {
		typeMappings = typeMappingsCleanupDirectory(dir, typeMappings);
		
		String[] brat2UimaTypeMappings = new String[typeMappings.length + 1];
		String[] uimaTypes = new String[typeMappings.length + 1];
		for (int ii=0; ii < typeMappings.length; ii++) {
			String bratAnnotation = typeMappings[ii][0];
			String uimaType = typeMappings[ii][1];
			brat2UimaTypeMappings[ii] = createBratJavaMapping(bratAnnotation, uimaType);
			uimaTypes[ii] = createAnnotationType(uimaType);
		}
		brat2UimaTypeMappings[brat2UimaTypeMappings.length - 1] = createBratJavaMapping(BRAT_UNKNOWN_LABEL, JAVA_UNKNOWN_TYPE);
		uimaTypes[uimaTypes.length - 1] = createAnnotationType(JAVA_UNKNOWN_TYPE);
		
		
		String strCorpusDir = dir.getAbsolutePath();
		String bratInputFiles = strCorpusDir + File.separator + "*" + BRAT_ANNOTATION_FILE_EXT;
		
		CollectionReaderDescription descr = CollectionReaderFactory.createReaderDescription(BratReader.class, 
				BratReader.PARAM_SOURCE_LOCATION, bratInputFiles,
				BratReader.PARAM_LANGUAGE, lang,
				BratReader.PARAM_TEXT_ANNOTATION_TYPE_MAPPINGS, brat2UimaTypeMappings,
				BratReader.PARAM_TEXT_ANNOTATION_TYPES, uimaTypes
			);		
		
		return descr;
	}

	public static CollectionReader bratDirReader(File dir) throws ResourceInitializationException {
		return bratDirReader(dir, "en", null);
	}
	
	public static CollectionReader bratDirReader(File dir, String[][] typeMappings) throws ResourceInitializationException {
		return bratDirReader(dir, "en", typeMappings);
	}
	
	public static CollectionReader bratDirReader(File dir, String lang) throws ResourceInitializationException {
		return bratDirReader(dir, lang, null);
	}

	public static CollectionReader bratDirReader(File dir, String lang, String[][] typeMappings) throws ResourceInitializationException {
		Logger tLogger = LogManager.getLogger("ca.nrc.nlp.brat.BratUtils.bratDirReader");
		if (tLogger.isTraceEnabled()) tLogger.trace("invoked with dir="+dir+", lang="+lang);
		
//		warnIfAnnFilesDoNotCorrespond(dir);
		
		CollectionReaderDescription descr = bratDirReaderDescr(dir, lang, typeMappings);
		CollectionReader reader = CollectionReaderFactory.createReader(descr);
		
		return reader;
	}

	private static void ensureMappingsArePair(String[][] typeMappings) {
		for (int ii=0; ii < typeMappings.length; ii++) {
			String[] mapping = typeMappings[ii];
			if (mapping.length != 2) {
				throw new IllegalArgumentException("Type mapping number "+ii+" was not a pair.\nValue was: ["+String.join(", ", mapping)+"]");
			}
		}
		
	}

	public static CollectionReader bratFileReader(File bratFile) throws ResourceInitializationException {
		return bratFileReader(bratFile, "en", null);
	}

	public static CollectionReader bratFileReader(File bratFile, String[][] typeMappings) throws ResourceInitializationException {
		return bratFileReader(bratFile, "en", typeMappings);
	}
		
	public static CollectionReader bratFileReader(File bratFile, String lang, String[][] typeMappings) throws ResourceInitializationException {
		Logger tLogger = LogManager.getLogger("ca.nrc.nlp.brat.BratUtils.bratFileReader");
		
		if (tLogger.isTraceEnabled()) tLogger.trace("invoked with bratFile="+bratFile+", lang="+lang);
		try {
			traceBratFilesFor("At start of method", bratFile, tLogger);
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		
		if (bratFile.isDirectory()) {
			throw new IllegalArgumentException("File "+bratFile+" is a directory.");
		}
		
		if (typeMappings == null) {
			typeMappings = new String[][] {};
		}

		File bratTxtFile = new File(bratFile.toString().replaceAll("\\."+BRAT_ANNOTATION_FILE_EXT+"$", "\\."+BRAT_TEXT_FILE_EXT));		
		File bratAnnFile = new File(bratFile.toString().replaceAll("\\."+BRAT_TEXT_FILE_EXT+"$", "\\."+BRAT_ANNOTATION_FILE_EXT));
		if(!bratAnnFile.toString().endsWith(BRAT_ANNOTATION_FILE_EXT)) {
			throw new IllegalArgumentException("Unrecognized file format: Only .txt and .ann files are accepted");
		}
		
		if (!bratAnnFile.exists()) {
			try {
				bratAnnFile.createNewFile();
			} catch (IOException e) {
				throw new ResourceInitializationException("Cannot create .ann file for .txt file: "+bratTxtFile, new Object[0], e);
			}
		}

		String bratOutputFolder = bratFile.getParent();
				
		String[] typeMappingStrings = new String[typeMappings.length+1];
		String[] uimaTypes = new String[typeMappings.length + 1]; 
		for (int ii=0; ii<typeMappings.length; ii++) {
			String bratAnnotation = typeMappings[ii][0];
			String uimaType = typeMappings[ii][1];
			typeMappingStrings[ii] = createBratJavaMapping(bratAnnotation, uimaType);
			uimaTypes[ii] = createAnnotationType(uimaType);
		}
		typeMappingStrings[typeMappingStrings.length - 1] = createBratJavaMapping(BRAT_UNKNOWN_LABEL, JAVA_UNKNOWN_TYPE);
		uimaTypes[uimaTypes.length - 1] = createAnnotationType(JAVA_UNKNOWN_TYPE);
		
		CollectionReader reader = CollectionReaderFactory.createReader(CollectionReaderFactory.createReaderDescription(BratReader.class, 
				BratReader.PARAM_SOURCE_LOCATION, bratAnnFile.toString(),
				BratReader.PARAM_LANGUAGE, lang,
				BratReader.PARAM_TEXT_ANNOTATION_TYPE_MAPPINGS, typeMappingStrings,
				BratReader.PARAM_TEXT_ANNOTATION_TYPES, uimaTypes
			));
		
		return reader;
	}

	public static AnalysisEngineDescription bratWriterDescr(File dir, String[][] typeMappings)  throws ResourceInitializationException{
		typeMappings = typeMappingsCleanupDirectory(dir, typeMappings);
		
		String[] uima2BratTypeMappings = new String[typeMappings.length];
		for (int ii=0; ii < typeMappings.length; ii++) {
			String bratAnnotation = typeMappings[ii][0];
			String uimaType = typeMappings[ii][1];
			uima2BratTypeMappings[ii] = createBratJavaMapping(uimaType, bratAnnotation);
		}
		
		String outputPath = dir.getAbsolutePath();
		
		return AnalysisEngineFactory.createEngineDescription(
				AnalysisEngineFactory.createEngineDescription(BratWriter.class, 
						BratWriter.PARAM_TARGET_LOCATION, outputPath, 
						BratWriter.PARAM_OVERWRITE, true, 
						BratWriter.PARAM_STRIP_EXTENSION, true,
						BratWriter.PARAM_EXCLUDE_TYPES,new String[] { 
								"de.tudarmstadt.ukp.dkpro.core.api.segmentation.type.Token",
								"de.tudarmstadt.ukp.dkpro.core.api.segmentation.type.Sentence",
								"de.tudarmstadt.ukp.dkpro.core.api.segmentation.type.Paragraph"
						},

						BratWriter.PARAM_ENABLE_TYPE_MAPPINGS, true,
						
						// If the class name is sufficient to be used as the
						// annotation type in the outputted .ann file, just remove this
						// PARAM_TYPE_MAPPINGS setting
						BratWriter.PARAM_TYPE_MAPPINGS, uima2BratTypeMappings
				)
				
//				// Clean up the way unknown types are written in .ann files
//				AnalysisEngineFactory.createEngineDescription(BratOutputCleaner.class,
//						BratOutputCleaner.PARAM_OUTPUT_DIR_PATH, outputPath
//				)
		);
	}
	
	private static String[][] typeMappingsCleanupDirectory(File dir, String[][] typeMappings){
		if (!dir.isDirectory()) {
			throw new IllegalArgumentException("File "+dir+" is not a directory.");
		}
		
		if (typeMappings == null) {
			typeMappings = new String[][] {};
		}
		ensureMappingsArePair(typeMappings);
		
		return typeMappings;
	}
	
	public static String txtFileContent(File bratFile) throws IOException {
		File bratTxtFile = txtFileFor(bratFile);
		String content = new String(Files.readAllBytes(bratTxtFile.toPath()));
		
		return content;
	}

	public static String annFileContent(File bratFile) throws IOException {
		File bratAnnFile = annFileFor(bratFile);
		String content = "[File does not exist]";
		if (bratAnnFile.exists()) {
			content = new String(Files.readAllBytes(bratAnnFile.toPath()));
		}
		
		return content;
	}

}
